#include "MPU9250.h"
#include "EEPROM.h"

#include <Eigen.h>
#include <ExtendedKalman.h>
#include <BaroSensor.h>
#include <elapsedMillis.h>

#include <struct_calibratedData.h>
#include <struct_euler.h>
#include <Quaternion.h>

s_calibratedData calibratedData;
s_euler orientation;
QuaternionClass quaternion;

//    KalmanClass kalmanPhi_, kalmanPsi_;
ExtendedKalmanClass EKF;

// an MPU-9250 object on SPI bus 0 with chip select 10
MPU9250 Imu(Wire, 0x68);
int status;

//baro
BaroSensorClass BaroSensor;

// a flag for when the MPU-9250 has new data
volatile int newData;
// EEPROM buffer and variables to load accel and mag bias 
// and scale factors from CalibrateMPU9250.ino
uint8_t EepromBuffer[48];
float axb, axs, ayb, ays, azb, azs;
float hxb, hxs, hyb, hys, hzb, hzs;

float ax, ay, az;
float gx, gy, gz;
float mx, my, mz;

float ypr[3]; // yaw pitch roll

// timers to measure performance
unsigned long lastUpdate, now;       // sample period expressed in milliseconds
float dt;

elapsedMillis sincePrint;

extern "C"{
  int _getpid(){ return -1;}
  int _kill(int pid, int sig){ return -1; }
  int _write(){return -1;}
}

void setup() {
  // serial to display data
  Serial.begin(115200);
  while(!Serial && millis()<2000) {}

  // start communication with IMU 
  status = Imu.begin();
  Wire.setClock(400000);
  if (status < 0) {
    Serial.println("IMU initialization unsuccessful");
    Serial.println("Check IMU wiring or try cycling power");
    Serial.print("Status: ");
    Serial.println(status);
    while(1) {}
  }
  Serial.println("IMU initialization successful");

    BaroSensor.begin();

  Serial.println("Baro sensor initialization successful");
  
  // load and set accel and mag bias and scale
  // factors from CalibrateMPU9250.ino 
  for (size_t i=0; i < sizeof(EepromBuffer); i++) {
    EepromBuffer[i] = EEPROM.read(i);
  }
  memcpy(&axb,EepromBuffer+0,4);
  memcpy(&axs,EepromBuffer+4,4);
  memcpy(&ayb,EepromBuffer+8,4);
  memcpy(&ays,EepromBuffer+12,4);
  memcpy(&azb,EepromBuffer+16,4);
  memcpy(&azs,EepromBuffer+20,4);
  memcpy(&hxb,EepromBuffer+24,4);
  memcpy(&hxs,EepromBuffer+28,4);
  memcpy(&hyb,EepromBuffer+32,4);
  memcpy(&hys,EepromBuffer+36,4);
  memcpy(&hzb,EepromBuffer+40,4);
  memcpy(&hzs,EepromBuffer+44,4);

  Serial.print( axb,6); Serial.print(", "); Serial.println(axs,6);
  Serial.print( ayb,6); Serial.print(", "); Serial.println(ays,6);
  Serial.print( azb,6); Serial.print(", "); Serial.println(azs,6);
  Serial.print( hxb,6); Serial.print(", "); Serial.println(hxs,6);
  Serial.print( hyb,6); Serial.print(", "); Serial.println(hys,6);
  Serial.print( hzb,6); Serial.print(", "); Serial.println(hzs,6);

  Imu.setAccelCalX(axb,axs);
  Imu.setAccelCalY(ayb,ays);
  Imu.setAccelCalZ(azb,azs);

  Imu.setMagCalX(hxb,hxs);
  Imu.setMagCalY(hyb,hys);
  Imu.setMagCalZ(hzb,hzs);


  // setting the accelerometer full scale range to +/-8G 
  Imu.setAccelRange(MPU9250::ACCEL_RANGE_4G);
  // setting the gyroscope full scale range to +/-500 deg/s
  Imu.setGyroRange(MPU9250::GYRO_RANGE_250DPS);


  // setting a 41 Hz DLPF bandwidth
  Imu.setDlpfBandwidth(MPU9250::DLPF_BANDWIDTH_41HZ);
  // setting SRD to 9 for a 100 Hz update rate
  // 1000/(1+SRD) = update rate
  // SRD = (1000/rate)-1,9
  Imu.setSrd(9);

  Imu.calibrateGyro();
  
  // enabling the data ready interrupt
  Imu.enableDataReadyInterrupt();
   //attaching the interrupt to microcontroller pin 1
   pinMode(14,INPUT);
   attachInterrupt(14,runFilter,RISING);

   lastUpdate = 0;
  
}

void loop() {  
  if (newData == 1) {
    newData = 0;
    now = millis();
    dt = ((now - lastUpdate) * 0.001);
    
    //if(dt/100 < 0.06) {
    quaternion = EKF.predict(&calibratedData, dt);
    //}
        
    // read the sensor
    // update the filter
    setDataStructure();

    //if(dt/100 < 0.06) {
    quaternion = EKF.predict(&calibratedData, dt);
    //}
        quaternion = EKF.update(&calibratedData, dt);
        lastUpdate = now; 
        
        getYawPitchRollDeg(&quaternion);
        //getYawPitchRoll180(&quaternion);
  }
} 

void getYawPitchRollDeg(QuaternionClass* q) {
  //float q[4]; // quaternion
  //float val[11];
  float gx, gy, gz; // estimated gravity direction
  //getQ(q, val);
  
  gx = 2 * (q->x*q->z - q->w*q->y);
  gy = 2 * (q->w*q->x + q->y*q->z);
  gz = q->w*q->w - q->x*q->x - q->y*q->y + q->z*q->z;
  
  ypr[0] = (180/pi) * atan2(2 * q->x * q->y - 2 * q->w * q->z, 2 * q->w*q->w + 2 * q->x * q->x - 1);
  ypr[1] = -(180/pi) * atan(gx / sqrt(gy*gy + gz*gz));
  ypr[2] = (180/pi) * atan(gy / sqrt(gx*gx + gz*gz));

  Serial.print(ypr[0]); Serial.print(","); Serial.print(ypr[1]);
  Serial.print(","); Serial.print(ypr[2]); Serial.println();
  /*  Serial.print(F("Orientation: "));
    Serial.print(ypr[0]);
    Serial.print(F(" "));
    Serial.print(ypr[1]);
    Serial.print(F(" "));
    Serial.print(ypr[2]);
    Serial.println(F(""));  */
}

void setDataStructure()
{
    Imu.readSensor();
    calibratedData.p = Imu.getGyroX_rads();
    calibratedData.q = Imu.getGyroY_rads();
    calibratedData.r = Imu.getGyroZ_rads();
    calibratedData.x = Imu.getAccelX_mss();
    calibratedData.y = Imu.getAccelY_mss();
    calibratedData.z = Imu.getAccelZ_mss();
    calibratedData.magx = Imu.getMagX_uT();
    calibratedData.magy = Imu.getMagY_uT();
    calibratedData.magz = Imu.getMagZ_uT();

    calibratedData.q = -calibratedData.q;
    
    calibratedData.temp = BaroSensor.getTemperature();
    calibratedData.pressure = BaroSensor.getPressure();
    calibratedData.altitude = ((pow(((1013.5) / calibratedData.pressure), 1/5.257) - 1.0) * (calibratedData.temp + 273.15)) / 0.0065;
}

void getYawPitchRoll180(QuaternionClass* q) {
  float gx, gy, gz;   // estimated gravity direction

  gx = 2 * (q->x*q->z - q->w*q->y);
  gy = 2 * (q->w*q->x + q->y*q->z);
  gz = q->w*q->w - q->x*q->x - q->y*q->y + q->z*q->z;

  //calculating yaw
  ypr[0] = (180./PI)*atan2(2 * q->x * q->y - 2 * q->w * q->z, 2 * q->w*q->w + 2 * q->x * q->x - 1); 
  //ypr[0] = val[9];
  if(ypr[0] > 180.) ypr[0] = ypr[0] - 360.;
  ypr[0] = ypr[0] * 0.0174532925;
  
  //calculating Pitch
  //Serial.print(gx); Serial.print("       "); Serial.print(gz); Serial.print("       ");
  if(gx > 0 && gz > 0) {
    ypr[1] = atan(gx / sqrt(gy*gy + gz*gz));
  } else if(gx > 0 && gz <= 0) {
    ypr[1] = PI - atan(gx / sqrt(gy*gy + gz*gz));
  } else if(gx < 0 && gz < 0) {
    ypr[1] = (-PI - atan(gx / sqrt(gy*gy + gz*gz)));
  } else  {
    ypr[1] =  atan(gx / sqrt(gy*gy + gz*gz));
  }
  
  //Calculating Roll1
        ypr[2] = atan(gy / sqrt(gx*gx + gz*gz));
        
        ypr[0]= ypr[0]*180./PI;
        ypr[1]= ypr[1]*180./PI;
        ypr[2]= ypr[2]*180./PI;
        
        //adjust pitch so its 0-360 deg
        ypr[1] = -ypr[1];
        if(ypr[1] < 0) ypr[1] = 360. + ypr[1];
   
    Serial.print(F("Orientation: "));
    Serial.print(ypr[0]);
    Serial.print(F(" "));
    Serial.print(ypr[1]);
    Serial.print(F(" "));
    Serial.print(ypr[2]);
    Serial.println(F(""));
}

void runFilter() {
  newData = 1;
}


